package aventi.epark.navigation.trilateration;

import org.apache.commons.math3.fitting.leastsquares.MultivariateJacobianFunction;
import org.apache.commons.math3.linear.Array2DRowRealMatrix;
import org.apache.commons.math3.linear.ArrayRealVector;
import org.apache.commons.math3.linear.RealMatrix;
import org.apache.commons.math3.linear.RealVector;
import org.apache.commons.math3.util.Pair;

/**
 * Created by espeland on 01.03.18.
 */

public class TrilaterationFunction implements MultivariateJacobianFunction {
    protected static final double epsilon = 1E-7;

    /**
     * Known coordinates of BLE Beacons
     */
    protected final double positions[][];

    /**
     * Euclidean distances mesaured between smartphone and BLE Beacons
     */
    protected final double distances[];

    public TrilaterationFunction(double positions[][], double distances[]) {

        if(positions.length < 3) {
            throw new IllegalArgumentException("Need at least three positions.");
        }

        if(positions.length != distances.length) {
            throw new IllegalArgumentException("The number of positions you provided, " + positions.length + ", does not match the number of distances, " + distances.length + ".");
        }

        // Bound distances to strictly positive domain
        for (int i = 0; i < distances.length; i++) {
            distances[i] = Math.max(distances[i], epsilon);
        }

        int positionDimension = positions[0].length;
        for (int i = 1; i < positions.length; i++) {
            if(positionDimension != positions[i].length) {
                throw new IllegalArgumentException("The dimensions (2D/3D) should be the same for all coordinates");
            }
        }

        this.positions = positions;
        this.distances = distances;
    }

    public final double[] getDistances() {
        return distances;
    }

    public final double[][] getPositions() {
        return positions;
    }

    /**
     * Calculate and return initialized Jacobian matrix for a given point
     *
     * Jacobian matrix, [i][j] at
     * J[i][0] = delta_[(x0-xi)^2 + (y0-yi)^2 - ri^2]/delta_[x0] at
     * J[i][1] = delta_[(x0-xi)^2 + (y0-yi)^2 - ri^2]/delta_[y0] partial derivative with respect to the parameters passed to value() method
     *
     * @param point for which to calculate the slope
     * @return Jacobian matrix for point
     */
    public RealMatrix jacobian(RealVector point) {
        double[] pointArray = point.toArray();

        double[][] jacobian = new double[distances.length][pointArray.length];
        for (int i = 0; i < jacobian.length; i++) {
            for (int j = 0; j < pointArray.length; j++) {
                jacobian[i][j] = 2 * pointArray[j] - 2 * positions[i][j];
            }
        }

        return new Array2DRowRealMatrix(jacobian);
    }

    @Override
    public Pair<RealVector, RealMatrix> value(RealVector point) {
        double[] pointArray = point.toArray();

        double[] resultPoint = new double[this.distances.length];

        // Calculate Least Squares
        for (int i = 0; i < resultPoint.length; i++) {
            resultPoint[i] = 0.0;
            // Calculate the sum and add to overall
            for (int j = 0; j < pointArray.length; j++) {
                resultPoint[i] += (pointArray[j] - this.getPositions()[i][j]) * (pointArray[j] - this.getPositions()[i][j]);
            }
            resultPoint[i] -= (this.getDistances()[i]) * (this.getDistances()[i]);
        }

        RealMatrix jacobian = jacobian(point);
        return new Pair<RealVector, RealMatrix>(new ArrayRealVector(resultPoint), jacobian);
    }
}
